module jp.ku6.windows.game.graphics;

import std.math;
import std.utf;
import std.c.stdio;
import std.c.stdlib;
import jp.ku6.windows.win32api.windef;
import jp.ku6.windows.win32api.winbase;
import jp.ku6.windows.win32api.winuser;
import jp.ku6.windows.win32api.wingdi;
import jp.ku6.windows.directx.d3d9types;
import jp.ku6.windows.directx.d3d9caps;
import jp.ku6.windows.directx.d3d9;
import jp.ku6.windows.game.application;

/++
 + グラフィックスクラス
 +/
class Graphics : WindowMessageListener
{
	/++
	 + 初期化処理
	 +/
	static
	void startup()
	{
		instance_ = new Graphics;
	}

	/++
	 + 終了処理
	 +/
	static
	void shutdown()
	{
		delete instance_;
	}

	@property
	{
		/++
		 + Singletonを取得
		 +/
		static
		Graphics instance()
		{
			return instance_;
		}

		IDirect3DDevice9 direct3DDevice()
		{
			return direct3DDevice_;
		}
	}

	/++
	 + 描画開始
	 +/
	void begin()
	{
		if (deviceLost_)
			return;

		direct3DDevice_.Clear(0, null, D3DCLEAR_TARGET, 0, 1.0f, 0);

		direct3DDevice_.BeginScene();
	}

	/++
	 + 描画終了
	 +/
	void end()
	{
		direct3DDevice_.EndScene();

		if (direct3DDevice_.Present(null, null, null, null) == D3DERR_DEVICELOST)
			deviceLost_ = true;
	}

	/++
	 + ディスプレイモードの変更
	 +/
	void resetDisplayMode(bool windowMode, int width, int height)
	{
		Window window = Application.instance.window;

		if (width == 0)
			width = window.clientWidth;
		if (height == 0)
			height = window.clientHeight;

		if (windowMode == windowMode_ && width == window.clientWidth && height == window.clientHeight)
			return;

		// フレームの有無
		window.frame = windowMode;

		// ウインドウのサイズ
		window.clientWidth = width;
		window.clientHeight = height;

		// ウインドウをリセット
		window.refresh();

		// プレゼントパラメータ設定
		setPresentParameters(windowMode, width, height);

		// デバイスをリセット
		if (windowMode)
		{
			direct3DDevice_.Reset(&presentParameters_);

			// ウインドウをセンタリングしておく
			window.centering();
			window.refresh();

			// 画面上で開いている他のウインドウを直す
		// m_WindowOperator.RestoreOthersWindow();
		}
		else
		{
			// 画面上で開いている他のウインドウの情報を保存
		// m_WindowOperator.GetOthersWindowInformation();

			direct3DDevice_.Reset(&presentParameters_);
		}
	}

	bool procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
	{
		if (deviceLost_)
		{
			switch (message) 
			{
			// ウインドウのアクティブ状態の変更時の処理
			case WM_ACTIVATE:
				if (LOWORD(wParam) == WA_INACTIVE)
					return true;
				break;

			// ウインドウのサイズが変更されたときの処理
			case WM_SIZE:
				if (wParam == SIZE_MINIMIZED)
					return true;
				break;

			// それ以外は何もしない
			default:
				return true;
			}

			// パラメータ再設定
			Window window = Application.instance.window;
			setPresentParameters(windowMode_, window.clientWidth, window.clientHeight);

			// リセット
			direct3DDevice_.Reset(&presentParameters_);

			deviceLost_ = false;
		}

		return true;
	}

	

private:

	// singleton
	static
	Graphics instance_;

	// 現在のスクリーンモード true=window  false=fullscreen
	bool windowMode_ = true;

	// デバイスロストフラグ
	bool deviceLost_;

	// d3d9.dll
	HMODULE d3d9Dll_;

	// Direct3D
	IDirect3D9 direct3D_;

	// Direct3D device
	IDirect3DDevice9 direct3DDevice_;

	// 描画設定
	D3DPRESENT_PARAMETERS presentParameters_;

	// ウインドウモードの設定
	D3DDISPLAYMODE windowDisplayMode_;

	// フルスクリーンモードの設定(フォーマットやリフレッシュレート違いで複数)
	D3DDISPLAYMODE[] fullScreenDisplayModes_;

	float[] cosTable_;

	// 他のウインドウを操作する
// CWindowOperator		m_WindowOperator;

	/+
	 + コンストラクタ
	 +/
	this()
	{
		// DLL読み込み
	    d3d9Dll_ = LoadLibrary("d3d9.dll\0".ptr);
	    if (d3d9Dll_ is null)
	        throw new GameError("Failed load d3d9.dll");
		scope(failure)
		{
			FreeLibrary(d3d9Dll_);
			d3d9Dll_ = null;
		}

		// 使用する関数のポインタを取得
		Direct3DCreate9Function Direct3DCreate9 = cast(Direct3DCreate9Function)GetProcAddress(d3d9Dll_, "Direct3DCreate9\0".ptr);
		if (Direct3DCreate9 is null)
			throw new GameError("Failed get Direct3DCreate9 function");

		// Direct3Dインターフェースを取得
		direct3D_ = Direct3DCreate9(D3D_SDK_VERSION);
		if (direct3D_ is null)
			throw new GameError("Failed get Direct3D interface");
		scope(failure)
		{
			direct3D_.Release();
			direct3D_ = null;
		}

		// ウインドウモード時のために現在の画面モードを取得
		if (FAILED(direct3D_.GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &windowDisplayMode_))) 
			throw new GameError("Failed get Adapter display mode");

		// フォーマットを指定して、フルスクリーン用の画面モードを取得
		AppendFullScreenDisplayMode(D3DFORMAT.D3DFMT_A8R8G8B8);
		AppendFullScreenDisplayMode(D3DFORMAT.D3DFMT_X8R8G8B8);


		Window window = Application.instance.window;

		// パラメータの設定
		setPresentParameters(windowMode_, window.clientWidth, window.clientHeight);

		// デバイスの作成 - T&L HAL
		if (FAILED(direct3D_.CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE.D3DDEVTYPE_HAL, window.handle, D3DCREATE_HARDWARE_VERTEXPROCESSING, &presentParameters_, &direct3DDevice_)))
		{
			// 失敗したので HAL で試行
			if (FAILED(direct3D_.CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE.D3DDEVTYPE_HAL, window.handle, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &presentParameters_, &direct3DDevice_)))
			{
				// 失敗したので REF で試行
				if (FAILED(direct3D_.CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE.D3DDEVTYPE_REF, window.handle, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &presentParameters_, &direct3DDevice_)))
				{
					direct3DDevice_ = null;
					throw new GameError("Failed get Direct3D device");
				}
			}
		}

		Application.instance.addWindowMessageListener(this);


		immutable double pai = 3.141592653589793238462643383279502884197169399375105820974944592;
		cosTable_ = new float[512];
		for (uint i; i < 512; ++i)
			cosTable_[i] = cast(float)std.math.cos(pai * cast(double)i / 256);

		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_DESTBLEND, D3DBLEND.D3DBLEND_INVSRCALPHA);

		// ZBufferをOFF
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_ZENABLE, D3DZBUFFERTYPE.D3DZB_FALSE);

		// FillModeは面で塗りつぶし
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_FILLMODE, D3DFILLMODE.D3DFILL_SOLID);

		// フラットシェーディングで
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_SHADEMODE, D3DSHADEMODE.D3DSHADE_FLAT);

		// 転送元ブレンディング係数を設定
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_SRCBLEND, D3DBLEND.D3DBLEND_SRCALPHA);

		// カリングを行わない
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_CULLMODE, D3DCULL.D3DCULL_NONE);

		// アルファテストはOFF
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_ALPHATESTENABLE, FALSE); 
	 
		// アルファブレンドは行う
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_ALPHABLENDENABLE, TRUE); 

		// フォグはOFF
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_FOGENABLE, FALSE); 

		// スペキュラはOFF
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_SPECULARENABLE, FALSE); 

		// ステンシル処理はOFF
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_STENCILENABLE, FALSE); 

		// プリミティブのクリッピングはON
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_CLIPPING, TRUE); 

		// ライトはOFF
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_LIGHTING, FALSE); 

		// アルファブレンドの設定
		direct3DDevice_.SetRenderState(D3DRENDERSTATETYPE.D3DRS_BLENDOP, D3DBLENDOP.D3DBLENDOP_ADD); 

		// 色の計算方法を指定
		direct3DDevice_.SetTextureStageState(0, D3DTEXTURESTAGESTATETYPE.D3DTSS_COLOROP, D3DTEXTUREOP.D3DTOP_MODULATE);

		// 色引数一つ目を設定
		direct3DDevice_.SetTextureStageState(0, D3DTEXTURESTAGESTATETYPE.D3DTSS_COLORARG1, D3DTA_TEXTURE);

		// 色引数二つ目を設定
		direct3DDevice_.SetTextureStageState(0, D3DTEXTURESTAGESTATETYPE.D3DTSS_COLORARG2, D3DTA_DIFFUSE);

		// アルファ値の計算方法を指定
		direct3DDevice_.SetTextureStageState(0, D3DTEXTURESTAGESTATETYPE.D3DTSS_ALPHAOP, D3DTEXTUREOP.D3DTOP_MODULATE);

		// アルファ値引数一つ目を設定
		direct3DDevice_.SetTextureStageState(0, D3DTEXTURESTAGESTATETYPE.D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

		// アルファ値引数二つ目を設定
		direct3DDevice_.SetTextureStageState(0, D3DTEXTURESTAGESTATETYPE.D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);

		// U座標テクスチャアドレッシング方法の設定
		direct3DDevice_.SetSamplerState(0, D3DSAMPLERSTATETYPE.D3DSAMP_ADDRESSU, D3DTEXTUREADDRESS.D3DTADDRESS_CLAMP);

		// V座標テクスチャアドレッシング方法の設定
		direct3DDevice_.SetSamplerState(0, D3DSAMPLERSTATETYPE.D3DSAMP_ADDRESSV, D3DTEXTUREADDRESS.D3DTADDRESS_CLAMP);

		// 拡大フィルタを設定
		direct3DDevice_.SetSamplerState(0, D3DSAMPLERSTATETYPE.D3DSAMP_MAGFILTER, D3DTEXTUREFILTERTYPE.D3DTEXF_POINT);

		// 縮小フィルタを設定
		direct3DDevice_.SetSamplerState(0, D3DSAMPLERSTATETYPE.D3DSAMP_MINFILTER, D3DTEXTUREFILTERTYPE.D3DTEXF_POINT);

		// 頂点フォーマットセット
		direct3DDevice_.SetVertexShader(null);
		direct3DDevice_.SetFVF(D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1);
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		Application.instance.removeWindowMessageListener(this);

		if (direct3DDevice_ !is null)
			direct3DDevice_.Release();

		if (direct3D_ !is null)
			direct3D_.Release();

		if (d3d9Dll_ !is null)
			FreeLibrary(d3d9Dll_);
	}

	/+
	 + Present parameterを設定
	 +/
	void setPresentParameters(bool windowMode, int width, int height)
	{
		(cast(ubyte*)&presentParameters_)[0..presentParameters_.sizeof] = 0; 

		windowMode_ = windowMode;
		if (windowMode)
		{
			presentParameters_.Windowed = TRUE;
			presentParameters_.BackBufferWidth = 0;
			presentParameters_.BackBufferHeight = 0;
			presentParameters_.BackBufferFormat = windowDisplayMode_.Format;
			presentParameters_.BackBufferCount = 1;
			presentParameters_.MultiSampleType = D3DMULTISAMPLE_TYPE.D3DMULTISAMPLE_NONE;
			presentParameters_.hDeviceWindow = Application.instance.window.handle;
			presentParameters_.EnableAutoDepthStencil = TRUE;
			presentParameters_.AutoDepthStencilFormat = D3DFORMAT.D3DFMT_D16;
			presentParameters_.SwapEffect = D3DSWAPEFFECT.D3DSWAPEFFECT_DISCARD;
			presentParameters_.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
		}
		else
		{
			// 最適なディスプレイモードを検索
			foreach (mode; fullScreenDisplayModes_)
			{
				if (mode.Width == width && mode.Height == height)
				{
					presentParameters_.Windowed = FALSE;
					presentParameters_.BackBufferWidth = mode.Width;
					presentParameters_.BackBufferHeight = mode.Height;
					presentParameters_.BackBufferFormat = mode.Format;
					presentParameters_.FullScreen_RefreshRateInHz = mode.RefreshRate;
					presentParameters_.BackBufferCount = 1;
					presentParameters_.MultiSampleType = D3DMULTISAMPLE_TYPE.D3DMULTISAMPLE_NONE;
					presentParameters_.hDeviceWindow = Application.instance.window.handle;
					presentParameters_.EnableAutoDepthStencil = TRUE;
					presentParameters_.AutoDepthStencilFormat = D3DFORMAT.D3DFMT_D16;
					presentParameters_.SwapEffect = D3DSWAPEFFECT.D3DSWAPEFFECT_DISCARD;
					presentParameters_.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;

					return;
				}
			}
		}
	}

	/+
	 + 色フォーマットを指定して、ディスプレイモードを取得する
	 +/
	void AppendFullScreenDisplayMode(D3DFORMAT format)
	{
		// アダプタが使用可能なフォーマットの数を取得する
		uint adapterModeCount = direct3D_.GetAdapterModeCount(D3DADAPTER_DEFAULT, format);

		int adapterModeCursor = 0;
		D3DDISPLAYMODE[] displayModes = new D3DDISPLAYMODE[adapterModeCount];

		for (uint i; i < adapterModeCount; ++i)
		{
			if (direct3D_.EnumAdapterModes(D3DADAPTER_DEFAULT, format, i, &displayModes[adapterModeCursor] ) == D3D_OK)
				++adapterModeCursor;	// 取得した数をインクリメント
		}

		// リフレッシュレートで昇順にソートする
		for (int i; i < adapterModeCursor - 1; ++i)
		{
			for (int j; j < adapterModeCursor - 1 - i; ++j)
			{
				if (displayModes[j].RefreshRate > displayModes[j + 1].RefreshRate)
				{
					D3DDISPLAYMODE tmp = displayModes[j];
					displayModes[j] = displayModes[j + 1];
					displayModes[j + 1] = tmp;
				}
			}
		}

		fullScreenDisplayModes_ ~= displayModes;
	}

	/+
	 + sinの値を取得
	 +
	 + Sprite用。何処に置くか迷ったのでとりあえずこのへんに・・・
	 +/
	float sin(uint angle)
	{
		return cosTable_[(384 + angle) & 0x000001FF];
	}

	/+
	 + cosの値を取得
	 +/
	float cos(uint angle)
	{
		return cosTable_[angle & 0x000001FF];
	}
}

/++
 + ピクセル画像データファイル
 +/
interface SurfaceSource
{
	@property
	{
		/// 画像の幅
		uint width();

		/// 画像の高さ
		uint height();
	}

	/++
	 + ファイルをオープン
	 +
	 + Returns:
	 +  成功した場合にtrueを返す
	 +/
	bool open();

	/++
	 + ファイルをクローズ
	 +/
	void close();

	/++
	 + データをメモリにロード
	 +
	 + Params:
	 +  buffer = 展開先メモリのポインタ
	 +  width = 展開先メモリ領域の幅
	 +  height = 展開先メモリ領域の高さ
	 +/
	void read(uint* buffer, uint width, uint height);
}

/++
 + ピクセル画像データ
 +/
class Surface
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
		// Graphicsの初期化を確認
		if (Graphics.instance is null)
			throw new GameError("Graphics is not initialized");
	}

	/++
	 + コンストラクタ
	 +
	 + 予めサイズを指定して領域を確保する
	 +
	 + Params:
	 +  width = 幅
	 +  height = 高さ
	 +/
	this(int width, int height)
	{
		this();
		resize(width, height);
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  source = データソース
	 +/
	this(SurfaceSource source)
	{
		load(source);
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		if (texture_ !is null)
		{
			texture_.Release();
			texture_ = null;
		}

		width_ = 0;
		height_ = 0;
	}

	@property
	{
		/// 幅
		const
		int width()
		{
			return width_;
		}

		/// 高さ
		const
		int height()
		{
			return height_;
		}

		IDirect3DTexture9 texture()
		{
			return texture_;
		}
	}

	/++
	 + 領域を再確保する
	 +
	 + 既にデータが書き込まれていた場合、データは消失する
	 +
	 + Params:
	 +  width = 幅
	 +  height = 高さ
	 +/
	void resize(int width, int height)
	{
		if (texture_ !is null)
		{
			texture_.Release();
			texture_ = null;
		}

		width_ = 0;
		height_ = 0;

		// テクスチャのサイズ設定
		int size;

		size = 1;
		for (;;)
		{
			if (size >= width)
			{
				width = size;
				break;
			}
			size *= 2;
		}

		size = 1;
		for (;;)
		{
			if (size >= height)
			{
				height = size;
				break;
			}
			size *= 2;
		}

		// アルファ付き32bitのフォーマットで作成
		if (Graphics.instance.direct3DDevice_.CreateTexture(width, height, 1, 0, D3DFORMAT.D3DFMT_A8R8G8B8, D3DPOOL.D3DPOOL_MANAGED, &texture_, null) != D3D_OK)
		{
			texture_ = null;
			return;
		}

		width_ = width;
		height_ = height;
	}

	/++
	 + ピクセルデータをロードする
	 +
	 + Params:
	 +  source = データソース
	 +/
	void load(SurfaceSource source)
	{
		if (source is null || !source.open())
			return;

		if (width_ < source.width || height_ < source.width || texture_ is null)
			resize(source.width, source.height);

		D3DLOCKED_RECT lockedRect;
		if (texture_.LockRect(0, &lockedRect, null, 0) == D3D_OK)
		{
			source.read(cast(uint*)lockedRect.pBits, width_, height_);

			texture_.UnlockRect(0);
		}

		source.close();
	}

private:

	// テクスチャ幅
	int width_;

	// テクスチャ高さ
	int height_;

	// テクスチャインターフェース
	IDirect3DTexture9 texture_;
}

/++
 + Surface内の矩形範囲
 +/
struct Pattern
{
	/// 左上点のサーフェス上でのX座標
	int u;

	/// 左上点のサーフェス上でのY座標
	int v;

	/// 幅
	int width;

	/// 高さ
	int height;

	/// サーフェス
	Surface surface;
}

/++
 + Patternの描画
 +/
struct Sprite
{
	enum DisplayType
	{
		NONE,      // 非表示
		NORMAL,    // 通常表示
		ADD_BLEND, // 加算合成
	}

	/// X軸表示座標
	float x = 0;

	/// Y軸表示座標
	float y = 0;

	/// X軸中心座標 この座標が表示座標で指定した点に来る 拡大・縮小・回転の中心にもなる
	float centerX = 0;

	/// Y軸中心座標 この座標が表示座標で指定した点に来る 拡大・縮小・回転の中心にもなる
	float centerY = 0;

	/// X軸せん断率
	float inclineX = 0;

	/// Y軸せん断率
	float inclineY = 0;

	/// X軸拡大率
	float scaleX = 1;

	/// Y軸拡大率
	float scaleY = 1;

	/// 回転角 一周を512とした時の値
	uint rotation = 0;

	/// 色
	uint color = 0xffffffff;

	// 表示形式
	DisplayType type = DisplayType.NORMAL;

	/// 表示する矩形範囲
	Pattern* pattern = null;

	/++
	 + 描画
	 +/
	void draw()
	{
		if (pattern is null || pattern.surface is null || pattern.surface.texture_ is null)
			return;

		IDirect3DDevice9 device = Graphics.instance.direct3DDevice_;

		switch (type)
		{
		case DisplayType.NONE:
			return;
		case DisplayType.NORMAL:
			device.SetRenderState(D3DRENDERSTATETYPE.D3DRS_DESTBLEND, D3DBLEND.D3DBLEND_INVSRCALPHA);
			break;
		case DisplayType.ADD_BLEND:
			device.SetRenderState(D3DRENDERSTATETYPE.D3DRS_DESTBLEND, D3DBLEND.D3DBLEND_ONE);
			break;
		default:
			break;
		}

		// テーブルからSin,Cosの値を取得
		float s = Graphics.instance.sin(rotation);
		float c = Graphics.instance.cos(rotation);

		// 回転行列×せん断行列作成
		float[2][2] mat;
		mat[0][0] = c - (inclineY * s);
		mat[0][1] = s + (inclineY * c);
		mat[1][0] = (inclineX * c) - s;
		mat[1][1] = c + (inclineX * s);

		// 拡大縮小行列を掛ける
		mat[0][0] *= scaleX;
		mat[0][1] *= scaleX;
		mat[1][0] *= scaleY;
		mat[1][1] *= scaleY;

		Vertex vertices[4];
		float lX;
		float lY;

		// 左上
		lX = -centerX;
		lY = -centerY;
		vertices[0].x = (mat[0][0] * lX) + (mat[1][0] * lY);
		vertices[0].y = (mat[0][1] * lX) + (mat[1][1] * lY);

		// 右上
		lX = cast(float)(pattern.width) - centerX;
		lY = -centerY;
		vertices[1].x = (mat[0][0] * lX) + (mat[1][0] * lY);
		vertices[1].y = (mat[0][1] * lX) + (mat[1][1] * lY);

		// 右下
		lX = cast(float)(pattern.width) - centerX;
		lY = cast(float)(pattern.height) - centerY;
		vertices[2].x = (mat[0][0] * lX) + (mat[1][0] * lY);
		vertices[2].y = (mat[0][1] * lX) + (mat[1][1] * lY);

		// 左下
		lX = -centerX;
		lY = cast(float)(pattern.height) - centerY;
		vertices[3].x = (mat[0][0] * lX) + (mat[1][0] * lY);
		vertices[3].y = (mat[0][1] * lX) + (mat[1][1] * lY);

		// スクリーン座標に変換
		lX = x - 0.5f;
		vertices[0].x += lX;
		vertices[1].x += lX;
		vertices[2].x += lX;
		vertices[3].x += lX;
		lY = y - 0.5f;
		vertices[0].y += lY;
		vertices[1].y += lY;
		vertices[2].y += lY;
		vertices[3].y += lY;

		// Z値
		vertices[0].z = 0;
		vertices[1].z = 0;
		vertices[2].z = 0;
		vertices[3].z = 0;

		// W値
		vertices[0].w = 1.0f;
		vertices[1].w = 1.0f;
		vertices[2].w = 1.0f;
		vertices[3].w = 1.0f;

		// 色
		vertices[0].color = color;
		vertices[1].color = color;
		vertices[2].color = color;
		vertices[3].color = color;

		lX = cast(float)(pattern.surface.width);
		// 左
		vertices[0].u = vertices[3].u = cast(float)(pattern.u) / lX;
		// 右
		vertices[1].u = vertices[2].u = cast(float)(pattern.u + pattern.width) / lX;

		lY = cast(float)(pattern.surface.height);
		// 上
		vertices[0].v = vertices[1].v = cast(float)(pattern.v) / lY;
		// 下
		vertices[2].v = vertices[3].v = cast(float)(pattern.v + pattern.height) / lY;

		// テクスチャをセット
		device.SetTexture(0, pattern.surface.texture_);

		// レンダリング
		device.DrawPrimitiveUP(D3DPRIMITIVETYPE.D3DPT_TRIANGLEFAN, 2, &vertices, Vertex.sizeof);
	}

private:

	static struct Vertex
	{
		float x,y,z,w;
		uint color;
		float u,v;
	}
}

/++
 + テクスチャにフォントを書き込むクラス
 +/
class Font : SurfaceSource
{
	// 太字
	bool bold;

	// 斜体
	bool italic;

	// 下線
	bool underline;

	// 打ち消し線
	bool strike;

	// フォントサイズ
	uint size;

	/++
	 + コンストラクタ
	 +/
	this()
	{
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  m = 表示メッセージ
	 +  f = フォント
	 +  s = フォントサイズ
	 +/
	this(string m, string f, uint s)
	{
		message = m;
		face = f;
		size = s;
	}

	@property
	{
		uint width()
		{
			return width_;
		}

		uint height()
		{
			return height_;
		}

		string message()
		{
			return std.utf.toUTF8(message_);
		}

		void message(string m)
		{
			message_ = std.utf.toUTF16(m);
		}

		string face()
		{
			return std.utf.toUTF8(face_[0 .. $-1]);
		}

		void face(string f)
		{
			face_ = std.utf.toUTF16(f) ~ 0;
		}
	}

	bool open()
	{
		close();

		if (message_ is null || face_ is null || size == 0)
			return false;

		int boldStyle = bold ? FW_BOLD : FW_REGULAR;

		// bool -> BOOL
		DWORD italicStyle = italic ? TRUE : FALSE;
		DWORD underlineStyle = underline ? TRUE : FALSE;
		DWORD strikeStyle = strike ? TRUE : FALSE;

		// フォントを作成
		fontHandle_ = CreateFontW(size,							// フォント高さ
									0,							// 文字幅
									0,							// テキストの角度	
									0,							// ベースラインとｘ軸との角度
									boldStyle,					// フォントの重さ（太さ）
									italicStyle,				// イタリック体
									underlineStyle,				// アンダーライン
									strikeStyle,				// 打ち消し線
									SHIFTJIS_CHARSET,			// 文字セット
									OUT_DEFAULT_PRECIS,			// 出力精度
									CLIP_DEFAULT_PRECIS,		// クリッピング精度
									DEFAULT_QUALITY,			// 出力品質
									DEFAULT_PITCH | FF_DONTCARE,// ピッチとファミリー
									face_.ptr);					// 書体名


		// メインウインドウのハンドルでデバイスコンテキストを取得
		deviceContextHandle_ = CreateCompatibleDC(null);

		// 新しく作ったフォントを選択
		oldFontHandle_ = cast(HFONT)SelectObject(deviceContextHandle_, fontHandle_);

		// フォント情報を取得
		TEXTMETRIC textMetric;
		GetTextMetrics(deviceContextHandle_, &textMetric);

		// 半角スペースと全角スペースの幅を取得
		SIZE stringSize;
		GetTextExtentPoint32W(deviceContextHandle_, " "w.ptr, 1, &stringSize);
		halfSpaceWidth_ = stringSize.cx;
		GetTextExtentPoint32W(deviceContextHandle_, "　"w.ptr, 1, &stringSize);
		fullSpaceWidth_ = stringSize.cx;

		// 回転行列作成
		MAT2 mat;
		mat.eM11.value = 1;
		mat.eM22.value = 1;

		// まず必要なバッファの最大サイズを取得 (ついでにフォントの高さも取得)
		uint maxBufferLength;
		uint line = 1;
		uint width;
		uint maxWidth;
		GLYPHMETRICS glyphMetrics;
		foreach (u; message_)
		{
			if (u == 0x000D)
				continue;

			if (u == 0x000A)
			{
				// 行数を足す
				++line;

				// 一行の最大長を更新
				if (width > maxWidth)
					maxWidth = width;

				// 一行の長さを初期化
				width = 0;
				continue;
			}

			if (u == 0x0020)
				width += halfSpaceWidth_;
			else if (u == 0x3000)
				width += fullSpaceWidth_;
			else
			{
				// 文字情報を取得
				uint length = GetGlyphOutlineW(deviceContextHandle_, u, GGO_GRAY8_BITMAP, &glyphMetrics, 0, null, &mat);
				if (length != GDI_ERROR)
				{
					// バッファ最大長を更新
					if (length > maxBufferLength)
						maxBufferLength = length;

					// 一文字分の横幅を足す
					width += glyphMetrics.gmCellIncX;
				}
			}
		}

		// 一行の最大長を更新
		if (width > maxWidth)
			maxWidth = width;

		// 書き出し領域の幅と高さを設定
		width_ = maxWidth;
		height_ = line * textMetric.tmHeight;

		fontHeight_ = textMetric.tmHeight;

		// バッファ長を設定
		maxBufferLength_ = maxBufferLength;

		return true;
	}

	void close()
	{
		if (deviceContextHandle_ !is null && oldFontHandle_ !is null)
		{
			SelectObject(deviceContextHandle_, oldFontHandle_);
			oldFontHandle_ = null;
		}

		if (fontHandle_ !is null)
		{
			DeleteObject(fontHandle_);
			fontHandle_ = null;
		}

		if (deviceContextHandle_ !is null)
		{
			DeleteObject(deviceContextHandle_);
			deviceContextHandle_ = null;
		}
	}

	void read(uint* buffer, uint width, uint height)
	{
		if (fontHandle_ is null)
			return;

		// バッファを確保
		ubyte[] fontBuffer = new ubyte[maxBufferLength_];

		TEXTMETRIC textMetric;
		GetTextMetrics(deviceContextHandle_, &textMetric);

		GLYPHMETRICS glyphMetrics;

		// 回転行列作成
		MAT2 mat;
		mat.eM11.value = 1;
		mat.eM22.value = 1;

		// テクスチャを一旦初期化
		for (uint i = 0; i < width * height; ++i)
			buffer[i] = 0;

		// 全ての文字を書き出し
		uint originX;
		uint originY;
		foreach (u; message_)
		{
			if (u == 0x0D)
				continue;

			if (u == 0x0A)
			{
				originY += fontHeight_;
				originX = 0;
				continue;
			}

			if (u == 0x0020)
				originX += halfSpaceWidth_;
			else if (u == 0x3000)
				originX += fullSpaceWidth_;
			else
			{
				// フォントデータを取得
				uint length = GetGlyphOutlineW(deviceContextHandle_, u, GGO_GRAY8_BITMAP, &glyphMetrics, fontBuffer.length, fontBuffer.ptr, &mat);
				if (length != GDI_ERROR)
				{
					// フォントピッチを計算
					uint fontPitch = (glyphMetrics.gmBlackBoxX + 3) & ~0x03;

					// 書き込み開始点の座標を計算
					uint startX = originX + glyphMetrics.gmptGlyphOrigin.x;
					uint startY = originY + (textMetric.tmAscent - glyphMetrics.gmptGlyphOrigin.y);

					ubyte* s = fontBuffer.ptr;
					for (uint y; y < glyphMetrics.gmBlackBoxY; ++y)
					{
						// 書き込み位置を取得
						uint* d = &buffer[(y + startY) * width + startX];

						// 一行分データを書き込む
						for (uint x; x < glyphMetrics.gmBlackBoxX; ++x)
						{
							// 取得した0 ～ 64の数値を0 ～ 255に広げてアルファ値とする
							// 残りの3Byteは白(FFFFFF)で埋める
							uint a = s[x];
							a = ((a << 26) - ((a << 18) & 0x01000000)) | 0x00ffffff;

							// 斜体にしたりすると前の文字で入力した領域とかぶることがあるので代入ではなく飽和加算する
							uint b = d[x];
							uint tmp = ((a & b) + (((a ^ b) >> 1) & 0x7f7f7f7f)) & 0x80808080;
							tmp = (tmp << 1) - (tmp >> 7);
							d[x] = ((a + b) - tmp) | tmp;
						}

						// 元データを次の行へ
						s += fontPitch;
					}

					// 次の文字へ移動
					originX += glyphMetrics.gmCellIncX;
				}
			}
		}
	}

private:

	// デバイスコンテキストハンドル
	HDC deviceContextHandle_;

	// フォント変更のためのフォントハンドルバックアップ
	HFONT oldFontHandle_;

	// セットするフォントハンドル
	HFONT fontHandle_;

	// 使用する文字の中で最大のバッファ
	uint maxBufferLength_;

	// フォントの高さ
	uint fontHeight_;

	// 全ての文字を書いたときの横幅
	uint width_;

	// 全ての文字を書いたときの高さ
	uint height_;

	// 半角スペースの幅
	uint halfSpaceWidth_;

	// 全角スペースの幅
	uint fullSpaceWidth_;

	// フォント
	wstring face_;

	// 文字列
	wstring message_;
}


/++
 + ビットマップファイル
 +/
class BitmapFile : SurfaceSource
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +  offset = ファイル先頭からのオフセット。1ファイルに複数のファイルをまとめた時に使用
	 +/
	this(string filePath, uint offset)
	{
		filePath_ = filePath ~ 0;
		fileOffset_ = offset;
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		close();
	}

	@property
	{
		uint width()
		{
			return width_;
		}

		uint height()
		{
			return height_;
		}

		string filePath()
		{
			return filePath_[0 .. $ - 1];
		}
	}

	bool open()
	{
		// ファイル自体をオープン
		file_ = std.c.stdio.fopen(filePath_.ptr, "rb0".ptr);
		if (file_ is null)
			return false;

		// アーカイブ内のファイル先頭位置へ移動
		if (fileOffset_ > 0)
			std.c.stdio.fseek(file_, fileOffset_, std.c.stdio.SEEK_SET);

		// 必要なところだけ読み出し
		ushort[15] data;
		std.c.stdio.fread(data.ptr, ushort.sizeof, data.length, file_);

		width_ = cast(uint)data[9];
		height_ = cast(uint)data[11];

		// TODO: BPPをチェックする処理

		return true;
	}

	void close()
	{
		std.c.stdio.fclose(file_);

		width_ = 0;
		height_ = 0;
	}

	void read(uint* buffer, uint width, uint height)
	{
		// TODO: 24bit以外の時の処理

		// 1行のピクセルサイズを取得
		uint pitch = (width_ * 24 / 8 + 3) & (~3);

		// ピクセルデータ用のメモリを確保
		ubyte* source = cast(ubyte*)std.c.stdlib.malloc(pitch * height_);
		scope(exit) std.c.stdlib.free(source);

		// ピクセルデータ読み出し
		std.c.stdio.fseek(file_, fileOffset_ + 54, std.c.stdio.SEEK_SET);
		std.c.stdio.fread(source, ubyte.sizeof, pitch * height_, file_);

		for (short y; y < height_; ++y)
		{
			uint* d = &buffer[y * width];
			ubyte* s = &source[(height_ - y - 1) * pitch];

			for (short x; x < width_; ++x)
			{
				uint color = 0xff000000 | (s[x * 3 + 2] << 16) | (s[x * 3 + 1] << 8) | s[x * 3];
				if (color == TRANSFER_COLOR)
					d[x] = 0;
				else
					d[x] = color;
			}
		}
	}

private:

	// 抜き色
	immutable uint TRANSFER_COLOR = 0xff00ff00;

	// 幅
	uint width_;

	// 高さ
	uint height_;

	// アーカイブ中での個別ファイルの位置
	uint fileOffset_;

	// ファイルポインタ
	FILE* file_;

	// ファイルパス
	string filePath_;
}

